import 'package:flutter/foundation.dart';
import 'package:flutter/widgets.dart';
import 'package:flutter/services.dart';
import 'package:flutter_live_data/flutter_live_data.dart';
import 'package:provider/provider.dart';
import 'commons/life_cycle_listener.dart';
import 'view.dart';
import 'vm.dart';
import 'commons/base.dart';

/// ViewController
class MVVMWidget<VM extends ViewModel> extends BaseStatefulWidget
    with LifeCycleListener
    implements LifeCycleObserver {
  MVVMWidget({
    @required Widget Function(BuildContext context, VM viewModel) view,
    @required VM Function(LifeCycleObserver lifeCycleObserver) viewModel,
    Key key,
  })  : this.viewBuilder = view,
        this.viewModelBuilder = viewModel,
        this._viewModelHolder = _ViewModelHolder(),
        super(key: key) {
    _viewModelHolder.viewModel = viewModelBuilder(this);
  }

  final Widget Function(BuildContext, VM) viewBuilder;
  final VM Function(LifeCycleObserver) viewModelBuilder;

  String get name => '{{sys.ui.View}}';

  VM get viewModel => _viewModelHolder.viewModel;
  final _ViewModelHolder<VM> _viewModelHolder;

  @override
  State<StatefulWidget> createState() => _VMState();

  @override
  void onDispose() {
    _liveData?.forEach((liveData) {
      liveData?.dispose();
    });
  }

  final List<LiveData> _liveData = <LiveData>[];

  @override
  void observeLiveData<T>(LiveData<T> lv) {
    _liveData.add(lv);
  }
}

/// View State
class _VMState<T extends ViewModel> extends State<MVVMWidget<T>> implements LifeCycleObserver {
  ViewModel get vm => widget.viewModel;
  final List<LiveData> _liveData = const <LiveData>[];

  @override
  void observeLiveData<T>(LiveData<T> liveData) {
    _liveData.add(liveData);
  }

  @override
  void dispose() {
    super.dispose();
    vm?.onDispose();
    widget.onDispose();
  }

  @override
  void initState() {
    super.initState();

    SystemChannels.lifecycle.setMessageHandler((message) async {
      if (message == '${AppLifecycleState.resumed}') {
        vm.onResume();
      } else if (message == '${AppLifecycleState.inactive}') {
        vm.onInactive();
      } else if (message == '${AppLifecycleState.paused}') {
        vm.onPause();
      } else if (message == '${AppLifecycleState.detached}') {
        vm.onDetach();
      }

      return message;
    });

    vm?.onInit();
  }

  @override
  Widget build(BuildContext context) {
    vm?.context = context;

    final view = widget.viewBuilder(context, vm);
    if (view is View) {
      Future.delayed(Duration(seconds: 0), () {
        view.widgetHolder.context = context;
        view.widgetHolder.widget = widget;
        view.onInit();
      });
    }

    vm?.onBuild();

    return vm.isDisposed
        ? ChangeNotifierProvider(
            create: (context) => vm,
            child: view,
          )
        : ChangeNotifierProvider.value(
            value: vm,
            child: view,
          );
  }
}

class _ViewModelHolder<VM extends ViewModel> {
  VM viewModel;
}

abstract class HierarchyMVVMWidget<VM extends ViewModel> {
  final _ViewModelHolder<VM> controller = _ViewModelHolder();

  VM get viewModel => controller.viewModel;

  Widget build(BuildContext context) {
    var mvvm = render(context);
    controller.viewModel = mvvm.viewModel;
    return mvvm;
  }

  MVVMWidget render(BuildContext context);

  void plug(View view, [Symbol name]) {
    view.plug(this as Widget, name);
  }
}
